// Intended to use together with: Automated device for continuous stirring while sampling in liquid chromatography systems, by Omer Markovitch (omermar@gmail.com), Jim Ottelé, Obe Veldman and Sijbren Otto.
// Firmware by Obe Veldman
// version 0.9
// software developed with the arduino ide/compiler Arduino 1.8.9
// arduino libraries
// u8g2 graphics library
// pidlib library
//***************
// Error detection: low or high measured rpm value (compared to the user set rpm) during more than 60s is an error.
// If error detected than after next powerup the error is reported and the error must be cleared before normal operation, by pressing the button for 5 seconds

#include <Arduino.h>
#include <U8g2lib.h>
#include <Wire.h>
#include <PID_v1.h>
#include <EEPROM.h>

U8G2_SSD1306_128X64_NONAME_1_HW_I2C u8g2(U8G2_R0, /* reset=*/ U8X8_PIN_NONE);

// constants 
const int pwmPin =10;        // motor PWM output 
const int hallPin =9;        // motor speed input 
const int LedPin =13;       // speed indication led

const int PushButtonPin=5; //=4;  // pullup = 1, button pressed = 0 press encoder
const int VbatPin = 1;  // 100 k 10 k >> 10% bat

const int encoder0PinA =3;//   = 2;
const int encoder0PinB =4;// = 3;

//int encoder0Pos = 0;
int encoder0PinALast = LOW;
int n = LOW;
int val_encoder;
const int position_increase = -1; // +1 or -1 to reverse the direction of the encoder

int i;
const int maxspeedselect=15; // *100 rpm
const int minspeedselect=2; // *100 rpm

int speed =64;
int flash;
int speedselect;
int numberofflashes;
long int vbat;
long int vmot;
long int nexttime;
long int flashtime;
long int flashwaittime;
long int boosttime;
long int secondsrun;
long int displaytime;

const long int waittime = 4000; 

long int batpercentage;
void batcheck()
{
     vbat = 0;
     vbat=(analogRead(VbatPin)+analogRead(VbatPin)+analogRead(VbatPin)+analogRead(VbatPin))/4;
      batpercentage= (vbat - 342)*100/188;     //(vbat - 342)*100/188;
       batpercentage/=5; batpercentage*=5; 
      if (batpercentage<0) batpercentage=0;
      if (batpercentage>100) batpercentage=100;
}

// to start motor on low pwm (function is not in use)
void motorboost()
{
    if (boosttime>millis()) analogWrite(pwmPin, 125); else analogWrite(pwmPin, speed);
}

volatile int encoder0Pos = 0;

void encodersetup() {
  pinMode(encoder0PinA, INPUT_PULLUP);
  pinMode(encoder0PinB, INPUT_PULLUP);
  attachInterrupt(1, doEncoder, CHANGE);  // encoder pin on interrupt 1 - pin 3
}

long int decoderwaittime=0;

void doEncoder() {
  /* If pinA and pinB are both high or both low, it is spinning
     forward. If they're different, it's going backward.

     For more information on speeding up this process, see
     [Reference/PortManipulation], specifically the PIND register.
  */
  if (millis()>decoderwaittime){
  if (digitalRead(encoder0PinA) == digitalRead(encoder0PinB)) {
    encoder0Pos=encoder0Pos-position_increase;
  } else {
    encoder0Pos=encoder0Pos+position_increase;
  }
  decoderwaittime=millis()+50;
  displaytime=millis()+4000; // 
  }
}

long int revtime,revtime0,revtimehalf, revtime1,revtime2,revtime3,revtime4;

void measurespeed()
{ int error;
const int timeout0=400;
  long int timeoutval, timezero, timeone;
 timeoutval=millis()+timeout0; error=0;
  revtime=0;
  do   // sync
  if (millis()>timeoutval) {error=1; break;} 
  while (digitalRead(hallPin) == LOW);
  //measurement
  timeoutval=millis()+timeout0;
  do   // 
  if (millis()>timeoutval) {error=1; break;} 
  while (digitalRead(hallPin) == HIGH);
   timeoutval=millis()+timeout0;
  timezero=micros();
  delay(15); // filter hall trigger
  do
  if (millis()>timeoutval) {error=1; break;} 
  while ((digitalRead(hallPin) == LOW));
  revtimehalf=micros();
   timeoutval=millis()+timeout0;
  delay(15); // filter hall trigger
  
  do   // 
  if (millis()>timeoutval) {error=1; break;}
  while ((digitalRead(hallPin) == HIGH));

  timeone=micros();
  revtime0=(timeone-timezero)/1000;
  if (error ==0)  { if ((revtime0 < 1.8* revtime1) || (revtime0> 2.2*revtime1)) 
  revtime1=(timeone-timezero)/1000;  //micros to millis
  }  
  else revtime1=revtime2;
  
  revtime=revtime1; //+revtime2;  //+revtime3+revtime4;
  revtimehalf=(revtimehalf-timezero)/1000;
  if (error ==1)  revtime=0;

}

// PID control only PI control is used
// Define Variables we'll be connecting to
double Setpoint, Input, Output;

//Specify the links and initial tuning parameters
double Kp=0.2, Ki=0.07, Kd=0;   // 0.05, 0.1
PID myPID(&Input, &Output, &Setpoint, Kp, Ki, Kd, DIRECT);

void pidsetup()
{
  //initialize the variables we're linked to
  Setpoint = speedselect*100;

  //turn the PID on
  myPID.SetMode(AUTOMATIC);
  myPID.SetSampleTime(100);
}

void pidloop()

{  //Input=1;
  if (revtime!=0)  {Input = 60000/revtime;} else {Input=1;} 
  myPID.Compute();
  analogWrite(pwmPin, Output);
}

int speedselectmem=0;

void setup() {
  // set pins as PWM and led drive outputs
     pinMode(pwmPin, OUTPUT);
    analogWrite(pwmPin, 0);
    pinMode(LedPin, OUTPUT);
    pinMode(hallPin, INPUT_PULLUP);
    pinMode(PushButtonPin, INPUT_PULLUP);// set pin as button input:
    pinMode(VbatPin,INPUT);
    analogReference(INTERNAL);
    // use serial as extra status indication
    Serial.begin(9600); 
    Serial.println("Stirrer project January 2020");    
    Serial.println("version 0.901");    
  Serial.println("See publication");
  pinMode (encoder0PinA, INPUT_PULLUP);
  pinMode (encoder0PinB, INPUT_PULLUP);
  speedselect=EEPROM.read(speedselectmem); speed= speedselect*10;
  encodersetup();
  Wire.setClock(400000);
  nexttime=millis();
  flashtime=millis();
  numberofflashes=0;
  boosttime=0;
  displaytime=0;
  flashwaittime=0;
  u8g2.begin();
  u8g2.setPowerSave(0);
  u8g2.setFont(u8g2_font_courB18_tf);
   //delay(1000);
   u8g2.firstPage();
    do { 
    u8g2.setCursor(0,24);    
    u8g2.print("Stirrer");  
    u8g2.setCursor(0,58);    
    u8g2.print("ver 0.8");
    } 
    while (u8g2.nextPage());
    delay(2000);
   encoder0Pos=speedselect*2;
   if (speedselect>maxspeedselect) speedselect=maxspeedselect; if (speedselect<minspeedselect) speedselect=minspeedselect;
   errorsetup();
    pidsetup();
    myPID.SetOutputLimits(15, 255);
  delay(500);
  }

//eeprom memory
int errordetectedmem=10;
int errortypemem=11;
int batterylowmem=12;
int totalerrorminmem=13;
int totalerrorhoursmem=14;
int errortime10minmem=15;
int errortime10hoursmem=16;
int errortime20minmem=17;
int errortime20hoursmem=18;
int errortime30minmem=19;
int errortime30hoursmem=20;
int errortime40minmem=21;
int errortime40hoursmem=22;

int totalruntime=30;
int lastruntime = 34;
int errordetected, acceptedspeederror, acceptedspeederrordetected, acceptedbatteryerrordetected;
int errortype;
long int time_nospeederror,time_nobatteryerror, errorminutes,errorhours;
long int errorcheckperiod= 5000;
long int max_errorperiod= 60000;
long int nexterrorchecktime = 0;
long int totalerrortime = 0;
long int errortime40=0;    
long int errortime30=0;
long int errortime20=0;
long int errortime10=0;

void errorEEPROMsave(int mem, long int val)
{ 
  val=val/60000; //to min or hours
  if (val<=120) { EEPROM.update(mem,val );  Serial.print("error min");Serial.print(int(val),DEC);} 
  if (val>=120) {val = val /60; EEPROM.update(mem+1,val );Serial.print("error hrs");Serial.print(int(val),DEC);}

}

void errorcheck()
{ int deltaspeed;
  if (nexterrorchecktime<millis())
  {
    nexterrorchecktime=millis()+errorcheckperiod;
  deltaspeed=speedselect*100 - 60000/revtime;
  errordetected = 0;
  if (deltaspeed<0) deltaspeed= -1*deltaspeed; //absolute value
  if (deltaspeed > 0.4 * speedselect*100) {errortype=4; errordetected=1; errortime40=errortime40+ errorcheckperiod; errorEEPROMsave(errortime40minmem,errortime40);}
  else {
  if (deltaspeed > 0.3 * speedselect*100) {errortype=3; errordetected=1;errortime30=errortime30+ errorcheckperiod;errorEEPROMsave(errortime30minmem,errortime30);}
  else {
  if (deltaspeed > 0.2 * speedselect*100) {errortype=2; errordetected=1;errortime20=errortime20+ errorcheckperiod;errorEEPROMsave(errortime20minmem,errortime20);}
  else {
  if (deltaspeed > 0.1 * speedselect*100) { errortype=1; errordetected=1; errortime10=errortime10+ errorcheckperiod;errorEEPROMsave(errortime10minmem,errortime10);}
  }
  }
  }
  if (errordetected==0) time_nospeederror=millis();
    else 
    {totalerrortime=totalerrortime+errorcheckperiod;
       errorminutes=totalerrortime/60000; 
       if (errorminutes<120) {EEPROM.update(totalerrorminmem,errorminutes);}
        else {errorhours = errorminutes/60; EEPROM.update(totalerrorhoursmem,errorhours);}
  if (acceptedspeederrordetected!=0)
  {    Serial.print(int(errortype),DEC);Serial.print(" error detection, time: ");Serial.println(int((-time_nospeederror+millis())/1000),DEC);
  Serial.print(int(errorminutes),DEC);Serial.print(" minutes ");Serial.println(int(errorhours),DEC);
  
  }
  
  if (errordetected==1) 
  {  
  if ((time_nospeederror + max_errorperiod) <millis()) 
      { acceptedspeederrordetected=errordetected; 
        EEPROM.update(errordetectedmem,acceptedspeederrordetected);
        if (acceptedspeederror<errortype) acceptedspeederror=errortype; EEPROM.update(errortypemem,acceptedspeederror);
      }
   }
  if (batpercentage <= 5) 
  { 
    if ((time_nobatteryerror+max_errorperiod)<millis()) {acceptedbatteryerrordetected=1; EEPROM.update(batterylowmem,acceptedbatteryerrordetected);
    }
  }
  else { time_nobatteryerror=millis();}
  }
 }
}

void errorsetup()
{
  acceptedspeederror = EEPROM.read(errortypemem);
  acceptedspeederrordetected = EEPROM.read(errordetectedmem);
  acceptedbatteryerrordetected=EEPROM.read(batterylowmem);
  errorminutes=EEPROM.read(totalerrorminmem);
  errorhours = EEPROM.read(totalerrorhoursmem);
  
  do
  errorstatusanderase();
  while ((acceptedspeederrordetected!=0)||(acceptedbatteryerrordetected!=0));
  nexterrorchecktime=millis()+30000; // 30 s to stabilize rpm before errorcheck
  
}

void infodisplay()
{   u8g2.firstPage();
    do
    {
     pidloop();
     u8g2.setFont(u8g2_font_courB18_tf);
     u8g2.setCursor(0,58);
     if (revtime>0)
     { 
       u8g2.print(60000/revtime,DEC);
     }
     else { u8g2.print("wait");}
     u8g2.setCursor(70,58); 
     u8g2.print("rpm"); 
    if (displaytime<millis())
    {
         u8g2.setCursor(46,24); u8g2.print("%");   // 70
         u8g2.setCursor(0,24); u8g2.print(batpercentage); 
    u8g2.setFont(u8g2_font_open_iconic_embedded_2x_t);
     u8g2.setCursor(60,24);
    u8g2.print(char(64)); 
    if (acceptedspeederrordetected != 0)
    {  
       u8g2.setFont(u8g2_font_open_iconic_embedded_4x_t);
       u8g2.setCursor(86,32);//86
      u8g2.print(char(71)); 
    }
    }
    else
    {
     u8g2.setCursor(0,24);
     u8g2.print(speedselect*100,DEC);
     u8g2.setFont(u8g2_font_open_iconic_embedded_4x_t); 
     u8g2.setCursor(86,32);
     u8g2.print(char(66));
    }
     pidloop();
    }
    while (u8g2.nextPage() );
  
}  // end infodisplay()

void errorstatusanderase()
{
      u8g2.firstPage();
     do { 

     do {  u8g2.setFont(u8g2_font_courB10_tf);
           acceptedbatteryerrordetected=0; // batterycheck removed from reporting!!!!!!!!!!!
           if ((acceptedspeederrordetected==0)&&(acceptedbatteryerrordetected==0))
           {
           u8g2.setCursor(0,12);
           u8g2.print("Status: OK");
           }
           if (acceptedspeederrordetected!=0)
           { 
           u8g2.setCursor(0,12); u8g2.print("E40 ");
           errortime40=EEPROM.read(errortime40minmem);
           if (errortime40 <120){
             u8g2.setCursor(48,12);  u8g2.print(errortime40);
             u8g2.setCursor(96,12);  u8g2.print("min");
           }
           else {
             u8g2.setCursor(48,12); 
             errortime40=EEPROM.read(errortime40hoursmem); u8g2.print(errortime40); 
             u8g2.setCursor(96,12); u8g2.print("hr");
           }
           
           u8g2.setCursor(0,24); u8g2.print("E30 ");
           errortime30=EEPROM.read(errortime30minmem);
           if (errortime30 <120){
             u8g2.setCursor(48,24);  u8g2.print(errortime30);
             u8g2.setCursor(96,24);  u8g2.print("min");
           }
           else {
             u8g2.setCursor(48,24); 
             errortime30=EEPROM.read(errortime30hoursmem); u8g2.print(errortime30); 
             u8g2.setCursor(96,24); u8g2.print("hr");
           }
           
           u8g2.setCursor(0,36); u8g2.print("E20 ");
           errortime20=EEPROM.read(errortime20minmem);
           if (errortime20 <120){
             u8g2.setCursor(48,36);  u8g2.print(errortime20);
             u8g2.setCursor(96,36);  u8g2.print("min");
           }
           else {
             u8g2.setCursor(48,36); 
             errortime40=EEPROM.read(errortime20hoursmem); u8g2.print(errortime20); 
             u8g2.setCursor(96,36); u8g2.print("hr");
           }

           u8g2.setCursor(0,48); u8g2.print("E10 ");
           errortime10=EEPROM.read(errortime10minmem);
           if (errortime10 <120){
             u8g2.setCursor(48,48);  u8g2.print(errortime10);
             u8g2.setCursor(96,48);  u8g2.print("min");
           }
           else {
             u8g2.setCursor(48,48); 
             errortime40=EEPROM.read(errortime10hoursmem); u8g2.print(errortime20); 
             u8g2.setCursor(96,48); u8g2.print("hr");
           }

           u8g2.setCursor(0,63);
           u8g2.print("press to erase");
           
           }
          
        }
       while (u8g2.nextPage() );
      delay(4000);
      if (digitalRead(PushButtonPin)==0) 
      {
        EEPROM.update(errordetectedmem,0);
        EEPROM.update(errortypemem,0);
        EEPROM.update(batterylowmem,0);
        EEPROM.update(totalerrorminmem,0);
        EEPROM.update(totalerrorhoursmem,0);

        
        EEPROM.update(errortime10minmem,0);
        EEPROM.update(errortime10hoursmem,0);
        EEPROM.update(errortime20minmem,0);
        EEPROM.update(errortime20hoursmem,0);
        EEPROM.update(errortime30minmem,0);
        EEPROM.update(errortime30hoursmem,0);
        EEPROM.update(errortime40minmem,0);
        EEPROM.update(errortime40hoursmem,0);

        acceptedbatteryerrordetected=0;
        acceptedspeederrordetected=0;
        time_nospeederror=millis();
        time_nobatteryerror=millis();
        u8g2.firstPage();
        u8g2.setFont(u8g2_font_courB10_tf);
        u8g2.setCursor(0,12);
        u8g2.print("Status: OK");       
        }
      } while (digitalRead(PushButtonPin)==0);
}

void loop() {
 
      pidloop();
      val_encoder= encoder0Pos;
      if (val_encoder<2*minspeedselect) {encoder0Pos=2*minspeedselect; ;val_encoder=2*minspeedselect;}
      if (val_encoder>2*maxspeedselect) { encoder0Pos=2*maxspeedselect; val_encoder=2*maxspeedselect;}
      if (speedselect== val_encoder/2){;}
      else
      {speedselect = val_encoder/2;
      if (speedselect> maxspeedselect ) speedselect=maxspeedselect; 
      if (speedselect< minspeedselect ) speedselect=minspeedselect;
        EEPROM.update(speedselectmem,speedselect);
        Serial.println("save speed setting...");
        nexterrorchecktime=millis()+30000; // 30 s to stabilize rpm
      }
      
      measurespeed();
      pidloop();
      batcheck();
      errorcheck();
      Serial.println(int(Output),DEC);
     infodisplay();   

     Setpoint = speedselect*100;
     pidloop();
     if (digitalRead(PushButtonPin)==0) errorstatusanderase();
     // repeat loop
}
